import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.lang.reflect.InvocationTargetException;

public class DebugTCPTool implements ActionListener, CaretListener, FocusListener, DebugTCPUI {

	// State information
	DebugTCPState st;

	// frames and panels
	private JFrame mainFrame;
	private JPanel mainPanel;
	private JPanel configAndEntryPanel;
	private JPanel configPanel;
	private JPanel sendPanel;
	private JPanel textDisplayPanel;
	private JPanel hexDisplayPanel;
	private JPanel statusRowPanel;

	// Configuration
	private JRadioButton clientRadioButton;
	private JRadioButton serverRadioButton;
	private ButtonGroup clientServerButtonGroup;
	private JLabel remoteHostEntryLabel;
	private JTextField remoteHostEntry;
	private JLabel portEntryLabel;
	private JTextField portEntry;
	private JButton connectButton;
	private JButton disconnectButton;
	private JLabel lineTerminatorEntryLabel;
	private JTextField lineTerminatorEntry;

	// Transmit
	private final static int numSendRows = 5;
	private JPanel[] sendRow;
	private JTextField[] sendTextEntry;
	private JButton[] sendTextButton;

	// Receive
	private boolean textDisplayAreaHasFocus;
	private JTextArea textDisplayArea;
	private JScrollPane textDisplayPane;
	private JTextArea hexDisplayArea;
	private JScrollPane hexDisplayPane;
	private Font displayFont;

	// Status row
	private int selectionStartInt;
	private int selectionEndInt;
	private JLabel statusLabel;
	private JButton swapSelection;
	private JTextField clientIPDisplay;
	private JTextField clientPortDisplay;
	private JButton clearButton;
	private JCheckBox autoClear;

	private DebugTCPThread netThread;

	// Miscellaneous
	private String errorMessage;

	// Constructor
	public DebugTCPTool() {
		st = new DebugTCPState();
		netThread = new DebugTCPThread(this,st);
		SwingUtilities.invokeLater( new Runnable() {
			public void run() {
				createAndShowGUI();
			}
		} );
	}
	
	// Create and show GUI
	private void createAndShowGUI() {
		// Create frame and panel
		JFrame.setDefaultLookAndFeelDecorated(true);
		mainFrame = new JFrame("TCP Debugging Tool by Michael Siegenthaler");
		mainPanel = new JPanel();
		mainPanel.setLayout(new BoxLayout(mainPanel,BoxLayout.PAGE_AXIS));
		configAndEntryPanel = new JPanel();
		configAndEntryPanel.setLayout(new GridLayout(1,0));
		configPanel = new JPanel();
		configPanel.setLayout(new GridLayout(0,2));
		configPanel.setBorder(BorderFactory.createCompoundBorder(
				       BorderFactory.createTitledBorder("Configuration"),
				       BorderFactory.createEmptyBorder(5,5,5,5)));
		sendPanel = new JPanel();
		sendPanel.setLayout(new GridLayout(0,1));
		sendPanel.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createTitledBorder("Transmit"),
					BorderFactory.createEmptyBorder(5,5,5,5)));
		textDisplayPanel = new JPanel();
		textDisplayPanel.setLayout(new BorderLayout());
		textDisplayPanel.setBorder(BorderFactory.createCompoundBorder(
				       BorderFactory.createTitledBorder("Received Text"),
				       BorderFactory.createEmptyBorder(5,5,5,5)));
		hexDisplayPanel = new JPanel();
		hexDisplayPanel.setLayout(new BorderLayout());
		hexDisplayPanel.setBorder(BorderFactory.createCompoundBorder(
				       BorderFactory.createTitledBorder("Received Data"),
				       BorderFactory.createEmptyBorder(5,5,5,5)));
		statusRowPanel = new JPanel();
		statusRowPanel.setLayout( new GridLayout(1,0) );
		statusRowPanel.setBorder( BorderFactory.createEmptyBorder(5,5,5,5) );

		// Configuration
		clientRadioButton = new JRadioButton("Client");
		clientRadioButton.setMnemonic('i');
		clientRadioButton.addActionListener(this);
		serverRadioButton = new JRadioButton("Server");
		serverRadioButton.setMnemonic('v');
		serverRadioButton.addActionListener(this);
		clientServerButtonGroup = new ButtonGroup();
		clientServerButtonGroup.add(clientRadioButton);
		clientServerButtonGroup.add(serverRadioButton);

		remoteHostEntryLabel = new JLabel("Remote host:");
		remoteHostEntry = new JTextField(15);
		remoteHostEntry.setText("127.0.0.1");
		portEntryLabel = new JLabel("Port:");
		portEntry = new JTextField(15);
		portEntry.setText("2400");
		connectButton = new JButton("Connect");
		connectButton.addActionListener(this);
		connectButton.setMnemonic('L');
		disconnectButton = new JButton("Close Connection");
		disconnectButton.addActionListener(this);
		disconnectButton.setMnemonic('o');
		lineTerminatorEntryLabel = new JLabel("Line terminator:");
		lineTerminatorEntry = new JTextField(15);
		lineTerminatorEntry.setText("\\r\\n");

		configPanel.add(clientRadioButton);
		configPanel.add(serverRadioButton);
		configPanel.add(remoteHostEntryLabel);
		configPanel.add(remoteHostEntry);
		configPanel.add(portEntryLabel);
		configPanel.add(portEntry);
		configPanel.add(connectButton);
		configPanel.add(disconnectButton);
		configPanel.add(lineTerminatorEntryLabel);
		configPanel.add(lineTerminatorEntry);

		// Transmit
		sendRow = new JPanel[numSendRows];
		sendTextEntry = new JTextField[numSendRows];
		sendTextButton = new JButton[numSendRows];
		for ( int i = 0 ; i < numSendRows ; i++) {
			sendRow[i] = new JPanel();
			sendRow[i].setLayout(new BoxLayout(sendRow[i],BoxLayout.LINE_AXIS));

			sendTextEntry[i] = new JTextField(20);
			sendTextEntry[i].addActionListener(this);

			sendTextButton[i] = new JButton("Send");
			sendTextButton[i].addActionListener(this);

			sendRow[i].add(sendTextEntry[i]);
			sendRow[i].add(sendTextButton[i]);

			sendPanel.add(sendRow[i]);
		}

		// Receive
		displayFont = new Font("Monospaced", Font.PLAIN, 12);
		textDisplayArea = new JTextArea(1,1);
		textDisplayArea.setFont(displayFont);
		textDisplayArea.setEditable(false);
		textDisplayArea.addFocusListener(this);
		textDisplayArea.addCaretListener(this);
		textDisplayPane = new JScrollPane(textDisplayArea);
		textDisplayPane.setPreferredSize(new Dimension(600,150));

		hexDisplayArea = new JTextArea(1,1);
		hexDisplayArea.setFont(displayFont);
		hexDisplayArea.setEditable(false);
		hexDisplayArea.setLineWrap(true);
		hexDisplayArea.setWrapStyleWord(true);
		hexDisplayArea.addFocusListener(this);
		hexDisplayArea.addCaretListener(this);
		hexDisplayPane = new JScrollPane(hexDisplayArea);
		hexDisplayPane.setPreferredSize(new Dimension(600,150));

		textDisplayPanel.add(textDisplayPane);
		hexDisplayPanel.add(hexDisplayPane);

		// Status row
		clearButton = new JButton("Clear");
		clearButton.addActionListener(this);
		clearButton.setMnemonic('r');

		swapSelection = new JButton("Swap");
		swapSelection.addActionListener(this);
		swapSelection.setMnemonic('w');

		clientIPDisplay = new JTextField(10);
		clientIPDisplay.setEditable(false);
		clientIPDisplay.setBackground(Color.white);
		clientPortDisplay = new JTextField(5);
		clientPortDisplay.setEditable(false);
		clientPortDisplay.setBackground(Color.white);

		autoClear = new JCheckBox("Auto clear");
		autoClear.setSelected(true);
		autoClear.setMnemonic('u');

		statusLabel = new JLabel("");
		updateStatusLabel(true);

		statusRowPanel.add(statusLabel);
		statusRowPanel.add(swapSelection);
		statusRowPanel.add(clientIPDisplay);
		statusRowPanel.add(clientPortDisplay);
		statusRowPanel.add(clearButton);
		statusRowPanel.add(autoClear);

		// Set up panel hierarchy
		configAndEntryPanel.add(configPanel);
		configAndEntryPanel.add(sendPanel);
		mainPanel.add(configAndEntryPanel);
		mainPanel.add(textDisplayPanel);
		mainPanel.add(hexDisplayPanel);
		mainPanel.add(statusRowPanel);

		// Put everything in a nice, consistent state
		updateGeneralState();

		// Show the gui
		mainFrame.getContentPane().add(mainPanel,BorderLayout.CENTER);
		mainFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		mainFrame.pack();
		mainFrame.setLocationRelativeTo(null);
		mainFrame.setVisible(true);

	}

	// Event handler
	public void actionPerformed(ActionEvent e) {

		Object src = e.getSource();

		if(src == connectButton) {
			if (st.actingAsServer) {
				// Is a server
				if (st.isListening) {
					// Is a server and is listening
					// Stop listening
					netThread.stopListening();
					st.isListening = false;
					updateGeneralState();
				} else {
					// Is a server and not listening
					// Create a thread that listens on a port
					try {
						int serverPort = Integer.parseInt(portEntry.getText());
						netThread.startListening(serverPort);
					} catch (NumberFormatException ex) {
						popupErrorMessage("Invalid port specified.");
					}
				}	
			} else {
				// Is a client
				// Open a connection to a remote host
				try {
					int serverPort = Integer.parseInt(portEntry.getText());
					st.isAttemptingToConnect = true; 
					updateGeneralState();
					updateStatusLabel();
					netThread.connectTo(remoteHostEntry.getText(),serverPort);
				} catch (NumberFormatException ex) {
					popupErrorMessage("Invalid port specified.");
				}
			}
		} else if (src == disconnectButton) {
			netThread.closeConnection();
		} else if (src == clearButton) {
			// Clear displays
			st.textReceived = "";
			st.hexReceived = "";
			textDisplayArea.setText("");
			hexDisplayArea.setText("");
			st.txLength = 0;
			st.rxLength = 0;
			updateStatusLabel(true);
			if(!st.isConnected) {
				st.remoteHost = "";
				st.remotePort = "";
				clientIPDisplay.setText("");
				clientPortDisplay.setText("");
			}
		} else if (src == swapSelection) {
			if (textDisplayAreaHasFocus) {
				hexDisplayArea.requestFocus();
				hexDisplayArea.select(selectionStartInt*3,selectionEndInt*3-1);
			} else {
				textDisplayArea.requestFocus();
				textDisplayArea.select(selectionStartInt,selectionEndInt);
			}
			updateStatusLabel(false);
		} else if (src == clientRadioButton) {
			st.actingAsServer = false;
			updateGeneralState();
		} else if (src == serverRadioButton) {
			st.actingAsServer = true;
			updateGeneralState();
		} else {
			for ( int i = 0 ; i < numSendRows ; i++) {
				if (src == sendTextEntry[i] || src == sendTextButton[i]) {
					if(st.isConnected) {
						// Send text
						ConvertIOData converter = new ConvertIOData(); 
						int bufLength;
						String str = sendTextEntry[i].getText()
							+ lineTerminatorEntry.getText();
						byte[] buf = str.getBytes();
						buf = converter.binarize(buf);
						bufLength = converter.lastBinarizeLength();
						st.txLength += bufLength;
						netThread.sendData(buf,bufLength);
						updateStatusLabel(true);
					}
				}
			}
		}

	}

	public void caretUpdate(CaretEvent e) {
		int start = e.getDot();
		int end = e.getMark();
		if ( e.getSource() == textDisplayArea ) {
			textDisplayAreaHasFocus = true;
		} else if ( e.getSource() == hexDisplayArea ) {
			start++;
			start /= 3;
			end++;
			end /= 3;
			textDisplayAreaHasFocus = false;
		}
		if(start < end) {
			selectionStartInt = start;
			selectionEndInt = end;
		} else {
			selectionStartInt = end;
			selectionEndInt = start;
		}
		if (st.displayAreaHasFocus) {
			SwingUtilities.invokeLater( new Runnable() {
				public void run() {
					updateStatusLabel();
				}
			} );
		}
	}

	public void focusGained(FocusEvent fe) {
		Object src = fe.getSource();
		if (src == textDisplayArea) {
			st.displayAreaHasFocus = true;
			updateStatusLabel();
		} else if (src == hexDisplayArea) {
			st.displayAreaHasFocus = true;
			updateStatusLabel();
		}
	}

	public void focusLost(FocusEvent fe) {
		Object src = fe.getSource();
		if (src == textDisplayArea) {
			if (!hexDisplayArea.isFocusOwner()) {
				st.displayAreaHasFocus = false;
				updateStatusLabel();
			}
		} else if (src == hexDisplayArea) {
			if (!textDisplayArea.isFocusOwner()) {
				st.displayAreaHasFocus = false;
				updateStatusLabel();
			}
		}
	}

	// Set a new connection state and call method to update buttons
	public void updateConnectionState() {
		try {
			SwingUtilities.invokeAndWait( new Runnable() {
				public void run() {
					updateGeneralState();
				}
			} );
		} catch (InterruptedException ex) {
		} catch (InvocationTargetException ex) {
		}
	}

	// Show an error message
	public void popupErrorMessage(String message) {
		errorMessage = message;
		SwingUtilities.invokeLater( new Runnable() {
			public void run() {
				JOptionPane.showMessageDialog(mainFrame,errorMessage,
					"Error",JOptionPane.ERROR_MESSAGE);
			}
		} );
	}	

	// Receive some data and display it
	public void displayReceivedData(byte[] data, int dataLength) {
		st.rxLength += dataLength;
		ConvertIOData converter = new ConvertIOData();
		converter.convert(data,dataLength);
		st.textReceived += converter.getText();
		st.hexReceived += converter.getHex();
		SwingUtilities.invokeLater( new Runnable() {
			public void run() {
				textDisplayArea.setText(st.textReceived);
				hexDisplayArea.setText(st.hexReceived);
				updateStatusLabel(true);
			}
		} );
	}

	// Display the remote address and port
	public void displayRemoteAddressAndPort(String remoteAddress, int remotePort) {
		st.remoteHost = remoteAddress;
		st.remotePort = String.valueOf(remotePort);
		SwingUtilities.invokeLater( new Runnable() {
			public void run() {
				clientIPDisplay.setText(st.remoteHost);
				clientPortDisplay.setText(st.remotePort);
			}
		} );
	}

	// Change UI to reflect whether program is running as a client or as a server
	private void updateGeneralState() {
		// Things that depend whether we are running as a client or as a server
		if (st.isNewConnection) {
			if (autoClear.isSelected()) {
				st.textReceived = "";
				st.hexReceived = "";
				textDisplayArea.setText("");
				hexDisplayArea.setText("");
				st.txLength = 0;
				st.rxLength = 0;
			}
			st.isNewConnection = false;
		}
		if (st.actingAsServer) {
			// Is a server
			serverRadioButton.setSelected(true);
			remoteHostEntry.setEnabled(false);
			remoteHostEntryLabel.setEnabled(false);
			if (st.isListening) {
				// Is a server and is listening
				connectButton.setText("Stop"); 
				connectButton.setMnemonic('p');
				connectButton.setEnabled(true);
				clientRadioButton.setEnabled(false);
				serverRadioButton.setEnabled(false);
			} else {
				// Is a server but is not listening
				if (st.isConnected) {
					// Is a server, not listening, but connected
					connectButton.setEnabled(false);
				} else {
					// Is a server, not listening, not connected
					connectButton.setText("Listen");
					connectButton.setMnemonic('L');
					connectButton.setEnabled(true);
					clientRadioButton.setEnabled(true);
					serverRadioButton.setEnabled(true);
				}	
			}
		} else {
			// Is a client
			clientRadioButton.setSelected(true);
			connectButton.setText("Connect");
			connectButton.setMnemonic('c');
			remoteHostEntry.setEnabled(true);
			remoteHostEntryLabel.setEnabled(true);
			if (st.isConnected) {
				// Is a client and is connected
				connectButton.setEnabled(false);
			} else {
				if (st.isAttemptingToConnect) {
					// Is a client and is attempting to connect
					connectButton.setEnabled(false);
					clientRadioButton.setEnabled(false);
					serverRadioButton.setEnabled(false);
				} else {
					// Is a client and is not connected
					connectButton.setEnabled(true);
					clientRadioButton.setEnabled(true);
					serverRadioButton.setEnabled(true);
				}
			}
		}
		// Things that do not depend on whether we are in client or server mode
		if (st.isConnected) {
			// Is connected
			disconnectButton.setEnabled(true);
			clientIPDisplay.setForeground(Color.black);
			clientPortDisplay.setForeground(Color.black);
			clientRadioButton.setEnabled(false);
			serverRadioButton.setEnabled(false);
		} else {
			// Is not connected
			disconnectButton.setEnabled(false);
			clientIPDisplay.setForeground(Color.gray);
			clientPortDisplay.setForeground(Color.gray);
		}	
		updateSendFieldState();
		updateStatusLabel();
	}

	// Enable or disable send buttons based on whether a connection exists or not
	private void updateSendFieldState() {
		for ( int i = 0 ; i < numSendRows ; i++ ) {
			if (st.isConnected) {
				sendTextButton[i].setEnabled(true);
			} else {
				sendTextButton[i].setEnabled(false);
			}
		}
	}

	private void updateStatusLabel() {
		if (st.displayAreaHasFocus) {
			updateStatusLabel(false);
		} else {
			updateStatusLabel(true);
		}
	}

	// Update the status label
	private void updateStatusLabel(boolean showRxTx) {
		if (st.isAttemptingToConnect) {
			statusLabel.setText("Connecting...");
		} else {
			if (showRxTx) {
				statusLabel.setText("Rx: " + st.rxLength + ", Tx: " + st.txLength);
			} else {
				statusLabel.setText("Selected: " + selectionStartInt + ", " + selectionEndInt);
			}
		}
	}

	// Main program
	public static void main(String[] args) {
		new DebugTCPTool();
	}

}
